"""
Rule class (daxfi package).

The representation of a firewall's rule.

  Copyright 2001, 2002 Davide Alberani <alberanid@libero.it>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

import StringIO, copy
from xml.dom import ext

from daxfi._rulesutils import *

__version__ = '0.8'


class RuleBase:
    """Class that define a firewall rule.

    A firewall rule is described with a DOM object.
    """
    converter = {}
    first_transformation = {}
    section_transformation = {}
    defaults = {}
    withLog = 1
    firewallName = 'unknown'

    def __init__(self, dom=None, rn='-1'):
        """Create a new rule.

        *dom* -- The DOM that describes the rule.

        *rn* -- The optional rule number (useful for 'insert'-like rules).
        """
        self.__rule_dom = None
        self.__modified_dom = None
        self.__command = None
        self.setDom(dom)
        self.setRuleNumber(rn)

    def getFirewallName(self):
        """Return the firewall name."""
        return self.firewallName

    def getAction(self):
        """Get the action name.

        Return the action name, one of the values in xml_commands.
        """
        if self.__rule_dom:
            if self.__rule_dom.documentElement:
                return self.__rule_dom.documentElement.nodeName
        return None

    def setDom(self, o):
        """Set the DOM."""
        self.__rule_dom = o
        self.__modified_dom = modify_dom(o, self.converter)

    def getModifiedDom(self):
        """Get the DOM as modified with the self.converter dictionary."""
        if self.__modified_dom:
            return self.__modified_dom.documentElement.cloneNode(1)
        return self.__modified_dom

    def getDom(self):
        """Get the DOM."""
        return self.__rule_dom

    def setRuleNumber(self, rn):
        """Set the rule number."""
        self.__rule_number = rn

    def getRuleNumber(self):
        """Get the rule number."""
        return str(self.__rule_number)

    def getXML(self):
        """Return a pretty XML string for this rule."""
        # Can't use cStringIO because it doesn't support unicode chars.
        dom = self.getDom()
        if not dom:
            return ''
        s = StringIO.StringIO()
        dom = dom.cloneNode(1)
        dom.doctype.systemId = DTD_NAME
        ext.PrettyPrint(dom, stream=s, encoding='ISO-8859-1')
        return s.getvalue()

    def __str__(self):
        """Print this rule."""
        return self.getRuleCommand()

    def __repr__(self):
        """The representation string for Rule objects."""
        return '<%s firewall rule at %s>' % (self.firewallName, id(self))

    def __cmp__(self, other):
        """Compare two rules."""
        if not isinstance(other, self.__class__):
            return -1
        return compare_dom(self.__rule_dom, other.__rule_dom)

    def __nonzero__(self):
        """The rule is 'true', if it generates a command."""
        if self.getRuleCommand():
            return 1
        return 0

    def copy(self):
        """Return a copy of itself."""
        return copy.deepcopy(self)

    def _getRuleCommand(self):
        """Return a string used to run the command for this rule.

        By default, just return the list of attributes of the
        DOM object.
        This method must be override in the 'Rule' class that
        inherits from this class.
        """
        opt = self.getModifiedDom()
        if not opt:
            sl_print_error('this rule is uninitialized')
            return ''
        act = self.getAction()
        if not act: return flat_print_attrs(opt)
        return act + ' ' + flat_print_attrs(opt)

    def getRuleCommand(self):
        """Return a string that is the command to run this rule."""
        if self.__command == None:
            self.__command = str(self._getRuleCommand())
        return self.__command


