"""
RuleBuilder class (daxfi package).

The RuleBuilder class, used to generate Rule objects.

  Copyright 2001, 2002 Davide Alberani <alberanid@libero.it>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

import re
from xml.dom.NodeFilter import NodeFilter
from xml.dom.NodeIterator import NodeIterator
from xml.dom import minidom
from xml.sax import SAXParseException

from daxfi._rulesutils import *
from daxfi._syslog import *

__version__ = '0.8'


class RuleBuilder:
    """Class used to build a set of Rule objects.

    The parseXMLString() method return a list of Rule objects
    for the given XML string.
    """
    first_transformation = {
                            xml_commands: {
                                'source-ip':        (transf_ip, CALL),
                                'destination-ip':   (transf_ip, CALL)
                            },
                            'protocol': {
                                'protocol':         (transf_proto, CALL),
                                'source-port':      (transf_port, CALL),
                                'destination-port': (transf_port, CALL),
                                'tcp-flags':        (transf_sort, CALL),
                                'icmp-type':        (transf_icmp, CALL)
                            },
                            'target': {
                                'reject-with':      (transf_icmp, CALL)
                            }
                        }
    section_transformation = {}

    def __init__(self, rc):
        """Initialize a RuleBuilder object; a Rule class must be provided."""
        self.clean()
        self.setRuleClass(rc)
        self.setFirewallName(rc.firewallName)

    def parseOnly(self, s):
        """Only parse a XML string; an hack used for performance reasons."""
        try:
            return minidom.parseString(s)
        except SAXParseException, e:
            sl_print_error('error parsing the XML: ' + str(e))
            return []

    def parseXMLString(self, s):
        """Parse a given XML string; return a list of Rule objects."""
        self.clean()
        rule_set = self.parseOnly(s)
        if not rule_set:
            return []
        # Normalization saves a lot of computation <g>.
        rule_set.normalize()
        # XXX: it seems that rule_set.doctype.systemId is NEVER set... :-|
        dtype = DTD_NAME
        if rule_set.doctype and rule_set.doctype.systemId:
            dtype = rule_set.doctype.systemId
        if dtype[-len(DTD_NAME):] != DTD_NAME:
            sl_print_warning('"' + DTD_NAME + '" DTD expected; ' +
                                '"' + dtype + '" found.')
        # Check processing instructions.
        ni = NodeIterator(rule_set,
                            NodeFilter.SHOW_PROCESSING_INSTRUCTION,
                            None, 0)
        cn = ni.nextNode()
        while cn:
            if cn.target == 'daxfi-process':
                # Search 'only-for'.
                if re.search('only-for\s*=\s*["\']\w+(,\w+)*["\']\s*',
                                cn.data):
                    sd = cn.data.find('\'')
                    if sd == -1:
                        sd = cn.data.find('"')
                    e = cn.data.rfind('\'')
                    if e == -1:
                        e = cn.data.rfind('"')
                    only_for = cn.data[sd+1:e]
                    only_for = only_for.split(',')
                    only_for = [x.strip() for x in only_for]
                    if only_for and self.__firewallName not in only_for:
                        return []
                # Search 'not-for'.
                elif re.search('not-for\s*=\s*["\']\w+(,\w+)*["\']\s*',
                                    cn.data):
                    sd = cn.data.find('\'')
                    if sd == -1:
                        sd = cn.data.find('"')
                    e = cn.data.rfind('\'')
                    if e == -1:
                        e = cn.data.rfind('"')
                    not_for = cn.data[sd+1:e]
                    not_for = not_for.split(',')
                    not_for = [x.strip() for x in not_for]
                    if not_for and self.__firewallName in not_for:
                        return []
                else:
                    sl_print_warning('unknown value for processing' + \
                                        ' instruction: "' + cn.data + '"')
            else:
                sl_print_warning('unknown processing instruction: "' + \
                                    cn.target + '"')
            cn = ni.nextNode()
        if rule_set.documentElement.nodeName in xml_commands:
            # We have a single rule.
            return self.createFromDOM(rule_set.cloneNode(1))
        elif rule_set.documentElement.nodeName != 'ruleset':
            sl_print_error('unknown main XML element: "' + \
                            rule_set.documentElement.nodeName + '"')
            return []
        # For every node in the XML file that is a firewall
        # command, generate a Rule.
        rlist = []
        domimp = minidom.DOMImplementation()
        ni = NodeIterator(rule_set, NodeFilter.SHOW_ELEMENT,
                            NodeFilterByElementName(xml_commands), 0)
        cn = ni.nextNode()
        while cn:
            doctype = domimp.createDocumentType(cn.nodeName, '', dtype)
            doc = domimp.createDocument(None, cn.nodeName, doctype)
            # We need to use a clone, otherwise the nextNode() method of
            # NodeIterator isn't happy.
            doc.replaceChild(cn.cloneNode(1), doc.documentElement)
            for i in self.createFromDOM(doc):
                rlist.append(i)
            cn = ni.nextNode()
        return rlist

    def setDom(self, o):
        """Set the DOM we're working on."""
        if o:
            self._rule_dom = self._modifyDom(o)
        else:
            self._rule_dom = None

    def getDom(self):
        """Get the DOM."""
        return self._rule_dom

    def _modifyDom(self, o):
        """Modify the attributes for this rule.

        This method is called by the setDom() method; it handles
        the 'first_transformation' and 'section_transformation' dictionaries
        to modify the DOM object so that it can be compare with the ones
        from other rules.
        """
        rlist = []
        ni = NodeIterator(o, NodeFilter.SHOW_ELEMENT,
                            NodeFilterByElementName(xml_targets +
                                                    xml_protocols +
                                                    xml_nat),
                            0)
        cn = ni.nextNode()
        # Internally tags like <tcp /> or <drop /> are never used:
        # they are converted to <protocol protocol="tcp" /> and
        # <target target="drop" />
        ## The same is true for <snat />, <masq />, etc. etc.
        while cn:
            nodetype = ''
            name = cn.nodeName
            if name in xml_targets and name != 'target':
                nodetype = 'target'
            elif name in xml_protocols and name != 'protocol':
                nodetype = 'protocol'
            elif name in xml_nat and name != 'nat':
                nodetype = 'nat'
            if nodetype:
                if cn not in rlist:
                    rlist.append(cn)
                i = o.createElement(nodetype)
                i.setAttributeNS('', nodetype, name)
                k = 0
                while k < len(cn.attributes):
                    i.setAttributeNS('',
                                        cn.attributes.item(k).name,
                                        cn.attributes.item(k).value)
                    k = k + 1
                cn.parentNode.appendChild(i)
            cn = ni.nextNode()
        for i in rlist:
            if i.parentNode:
                i.parentNode.removeChild(i)
        # Normalize attribute negation ('! value') and strip values.
        o = modify_dom(o, forEveryAttribute=transf_normalize)
        # Do default transformations (IP, ports, protocols, etc.)
        o = modify_dom(o, convAttributes=self.first_transformation,
                            convSections=self.section_transformation)
        # Do custom transformations for the given Rule class.
        o = modify_dom(o, convAttributes=self.__ruleClass.first_transformation,
                        convSections=self.__ruleClass.section_transformation)
        # Set defaults.
        if o:
            set_dom_defaults(o.documentElement, self.__ruleClass.defaults)
        return o

    def clean(self):
        """Reset the builder."""
        self.setDom(None)

    def createFromDOM(self, dom):
        """Create a set of rules from a given DOM."""
        self.setDom(dom)
        return self.getRules()

    def setRuleClass(self, rc):
        """An hack."""
        self.__ruleClass = rc

    def setFirewallName(self, fn):
        """Another hack."""
        self.__firewallName = fn

    def getRulesWithoutLog(self):
        """Return the rules for this DOM, for firewalls without
        a log option (they need a separated LOG rule, like iptables)."""
        dom = self.getDom()
        if not dom:
            return []
        doc0 = dom.cloneNode(1)
        r0 = doc0.documentElement
        rlist = []
        cmd = r0.nodeName
        rn = '-1'
        rn_dom = r0.getAttributeNS('', 'rule-number')
        if rn_dom:
            rn = rn_dom
            r0.removeAttributeNS('', 'rule-number')
        have_log = r0.getElementsByTagName('log')
        have_target = r0.getElementsByTagName('target') or \
                            r0.getElementsByTagName('accept') or \
                            r0.getElementsByTagName('drop') or \
                            r0.getElementsByTagName('reject')
        if have_log:
            if have_target:
                # We have both <target /> and <log />
                # r0/doc0 is the one without the log and with the target.
                # r1/doc1 is the one with only the <log /> tag.
                doc1 = doc0.cloneNode(1) 
                r1 = doc1.documentElement
                have_log[0].parentNode.removeChild(have_log[0])
                target_r1 = r1.getElementsByTagName('target') or \
                                r0.getElementsByTagName('accept') or \
                                r0.getElementsByTagName('drop') or \
                                r0.getElementsByTagName('reject')
                target_r1[0].parentNode.removeChild(target_r1[0])
                lim = r1.getElementsByTagName('limit')
                llim = len(lim)
                if llim == 2 and r0.getElementsByTagName('limit'):
                    r1.removeChild(lim[0])
                rlist = [self.__ruleClass(doc1, rn),
                            self.__ruleClass(doc0, rn)]   
            else:
                # We got the log, but not the target.
                rlist = [self.__ruleClass(doc0, rn)]
        else:
            # we don't have the log.
            rlist = [self.__ruleClass(doc0, rn)]

        return rlist

    def getRulesWithLog(self):
        """Return the rules for this DOM.

        Elaborate the current DOM and return a set of Rule object.
        For firewalls that can log and set rules with a single command.
        """
        doc = self.getDom()
        if not doc:
            return [] 
        r = doc.documentElement
        rn = '-1'
        if r.getAttributeNS('', 'rule-number'):
            rn = r.getAttributeNS('', 'rule-number')
            r.removeAttributeNS('', 'rule-number')
        return [self.__ruleClass(doc, rn)]

    def getRules(self):
        """Return the rules for this DOM."""
        if self.__ruleClass.withLog:
            return self.getRulesWithLog()
        else:
            return self.getRulesWithoutLog()


