"""
daxfi package.

This package is used to build DAXFi-based applications.

  Copyright 2001, 2002 Davide Alberani <alberanid@libero.it>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

import types, commands

from daxfi._exceptions import *
from daxfi._syslog import *
from daxfi.RuleBuilder import RuleBuilder

__all__ = ['Rule', 'RuleBuilder', 'firewalls']

__version__ = '0.8'


# - Main class

class Firewall:
    """The main DAXFi class.

    Through an instance of this class, you have complete control over
    your firewall.
    """
    def __init__(self, firewallBrand=None, substitutionDict={}):
        """Return an instance of the Firewall class.

        Initialize a Firewall object.

        *firewallBrand* -- a string or a list of strings amongst
                            which the firewall will be selected.

        *substitutionDict* -- the dictionary used to substitute some
                                predefined strings in XML files.
        """
        self.__ruleClass = None
        self.__ruleBuilder = None
        self.__fw_mod = None
        self.__fw_name = ''
        self.__fw_forced = 0

        self.setSubstitutionDict(substitutionDict)
        self._loadFirewall(brand=firewallBrand)

    def __repr__(self):
        """Return the representation for a Firewall object."""
        return '<%s firewall at %x>' % (self.getFirewallName(), id(self))

    def _loadFirewall(self, brand=None):
        """Detect the firewall, and import related modules."""
        from daxfi import firewalls
        if brand and type(brand) in (types.StringType, types.UnicodeType):
            brand = (brand,)
        # If not specified, we'll try any known firewall.
        if not brand:
            brand = firewalls.__all__
        mod = None
        for i in brand:
            try:
                mod = __import__('daxfi.firewalls.' + i,
                                    None, None, i)
            except ImportError, e:
                sl_print_error(str(e))
                continue
            funct = mod.__dict__.get('is_supported')
            if funct:
                if funct():
                    # The firewall is supported.
                    self.__fw_mod = mod
                    self.__fw_forced = 0
                elif mod and len(brand) == 1:
                    # A single firewall was specified, and it's not
                    # supported, so that we're forced to use it.
                    self.__fw_mod = mod
                    self.__fw_forced = 1
                break
        else:
            raise DetectFirewallError, 'unable to detect the firewall'
        del firewalls
        if not self.__fw_mod:
            raise DetectFirewallError, 'unable to detect the firewall'
        self.__ruleClass = self.__fw_mod.Rule
        self.__fw_name = self.__fw_mod.Rule.firewallName
        self.__ruleBuilder = RuleBuilder(self.__ruleClass)

    def _runCommand(self, cmd):
        """Execute a given command."""
        if not cmd:
            return 0
        (status, output) = commands.getstatusoutput(cmd)
        if status:
            sl_print_error('got status error "' + str(status) + \
                            '" executing "' + cmd + '"; the output was: "' + \
                            output + '"')
            return 0
        return 1

    def forceFirewall(self, firewallBrand):
        """Force the use of the given firewall."""
        self._loadFirewall(brand=firewallBrand)

    def isForced(self):
        """Return true if the used firewall was forced."""
        return self.__fw_forced
 
    def setSubstitutionDict(self, substitutionDict):
        """Set the substitution dictionary."""
        self.__subDict = substitutionDict

    def getSubstitutionDict(self):
        """Get the substitution dictionary."""
        return self.__subDict

    def getFirewallName(self):
        """Return the name of the running firewall."""
        return self.__fw_name

    def getRuleClass(self):
        """Return the class that represents a Rule for this firewall."""
        return self.__ruleClass

    def getRuleBuilder(self):
        """Return the RuleBuilder object used to create new rules."""
        return self.__ruleBuilder

    def _subDictSubstitution(self, s):
        """Substitute well known strings in the given XML string with
        user supplied data."""
        for k, v in self.__subDict.items():
            s = s.replace(k, v)
        return s

    def _parseXMLFile(self, fname):
        """Parse a XML file, return a list of rules."""
        file_content = ''
        try:
            f = open(fname, 'r')
            file_content = f.read()
            f.close()
        except IOError:
            sl_print_error('unable to read file "' + fname + '"')
        return self._parseXMLString(file_content)

    def _parseXMLString(self, s):
        """Parse a XML string, with substitution of the user supplied
        substituition dictionary."""
        return self.__ruleBuilder.parseXMLString(self._subDictSubstitution(s))

    def newRulesFromXMLFile(self, fileName):
        """Parse XML file.

        Parse a XML file, and return a list of rules.

        *fileName* -- the name of the file to parse.
        """
        return self._parseXMLFile(fileName)

    def newRulesFromXMLString(self, s):
        """New rules from XML strings.

        Parse a XML string and return a list of rules.
        """
        return self._parseXMLString(s)

    def _create(self,
            action='append',
            pos='-1',
            target='drop',
            chain='input',
            source='0.0.0.0/0.0.0.0',
            destination='0.0.0.0/0.0.0.0',
            interface='',
            proto='',
            sport='',
            dport='',
            log=0):
        """Create a set of rules with the given parameters.
        By default the chain is 'input' and the target is 'drop'.
        """
        p = ''
        if proto:
            p = ' protocol="%s" ' % (proto)
        if sport:
            p = p + ' source-port="%s"' % (sport)
        if dport:
            p = p + ' destination-port="%s"' % (dport)
        if p:
            p = '<protocol %s />' % (p)
        if log:
            p = p + '<log />'
        p = p + '<target target="%s" />' % (target)

        x = '<?xml version="1.0"?><%s chain="%s" source-ip="%s" ' \
                'destination-ip="%s" ' % (action, chain, source, destination)
        if pos != '-1':
            x = x + ' rule-number="%s" ' % (str(pos))
        if interface:
            x = x + ' interface="%s" ' % (interface)
        x = x + '>'
        p = '%s%s</%s>' % (x, p, action)
        return self.newRulesFromXMLString(p)

    def createNewRules(self, *args, **kw):
        """Create new rules with the given parameters."""
        return apply(self._create, args, kw)

    def runRules(self, l):
        """Execute a given rule or list of rules."""
        if type(l) not in (types.ListType, types.TupleType):
            return self._runCommand(l.getRuleCommand())
        r = 1
        for i in l:
            if not self._runCommand(i.getRuleCommand()):
                r = 0
        return r

    def runXMLFile(self, fileName):
        """Parse an XML file and run the resulting set of rules."""
        return self.runRules(self.newRulesFromXMLFile(fileName))

    def runXMLString(self, s):
        """Parse an XML string and run the resulting set of rules."""
        return self.runRules(self.newRulesFromXMLString(s))

    def runNewRules(self, *args, **kw):
        """Create and run new rules with the given parameters."""
        return self.runRules(apply(self.createNewRules, args, kw))

    def getRuleCommands(self, l):
        """Return a list of commands for the given rule or list of rules."""
        rlist = []
        if type(l) not in (types.ListType, types.TupleType):
            rlist.append(str(l))
        else:
            for i in l:
                rlist.append(str(i))
        return rlist

    def getXMLRules(self, l):
        """Return a list of XML strings for the given rule or list of rules."""
        rlist = []
        if type(l) not in (types.ListType, types.TupleType):
            rlist.append(l.getXML())
        else:
            for i in l:
                rlist.append(i.getXML())
        return rlist

    def listChains(self): 
        """Return the list of chains available for this firewall."""
        try:    
            return self.__fw_mod.list_chains()
        except exceptions.OSError, s:
            if self.__fw_forced:
                return []
            else:
                raise exceptions.OSError, s

    def listRules(self, chain):
        """Return the list of rules in the given chain."""
        return [self.__ruleClass(self.__ruleBuilder.parseOnly(i))
                    for i in self.listXMLRules(chain)]

    def listCommandRules(self, chain):
        """Return a list of commands for the given chain."""
        return [self.__ruleClass(
                        self.__ruleBuilder.parseOnly(i)).getRuleCommand()
                    for i in self.listXMLRules(chain)]

    def listXMLRules(self, chain):
        """Return a list of XML strings; one for every running rule."""
        try:
            return self.__fw_mod.list_rules(str(chain))
        except exceptions.OSError, s:
            if self.__fw_forced:
                return []
            else:
                raise exceptions.OSError, s

    def checkRule(self, r):
        """If a rule is actually running, return the rule number,
           otherwise -1."""
        chain = ''
        chain = r.getDom().documentElement.getAttributeNS('', 'chain')
        if not chain:
            if r.getDom().getElementsByTagName('nat'):
                chain = 'nat'
            else:
                return -1
        rlist = self.listRules(chain)
        i = 0
        for item in rlist:
            i = i + 1
            if r == item:
                return i
        return -1


class _SingletonFactory:
    """Create a singleton of a given class."""
    __baseClass = None
    __object = None
    def __init__(self, base_class):
        self.__baseClass = base_class

    def __call__(self, *args, **kw):
        if self.__object is None:
            self.__object = apply(self.__baseClass, args, kw)
        return self.__object


# A 'singletonized' version of the Firewall class.
SingletonFirewall = _SingletonFactory(Firewall)


