"""
_rulesutils module (daxfi package).

This module provide utilities for the RuleBuilder and Rule classes.

  Copyright 2001, 2002 Davide Alberani <alberanid@libero.it>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

import types, re, socket, imp, os

from xml.dom.NodeFilter import NodeFilter
from xml.dom.NodeIterator import NodeIterator
from xml.dom import Node
from daxfi._exceptions import *
from daxfi._syslog import *
from daxfi import iplib

__version__ = '0.8'


# --- Strings that are XML tags

# - These tuples are used by the XML parser to know the kind of tag it
# - musts expect.

# Names of the firewall commands as used in the XML.
xml_commands = ('append', 'delete', 'replace', 'insert', 'flush', 'set-policy')

# Supported protocols.
xml_protocols = ('tcp', 'udp', 'icmp', 'protocol')

# Known targets.
xml_targets = ('accept', 'reject', 'drop', 'target')

# NAT types.
xml_nat = ('snat', 'dnat', 'masq', 'redirect', 'nat')

# Misc tags.
xml_extensions = ('log', 'limit', 'state')

# The current DTD.
DTD_NAME = 'daxfifw_0.4.dtd'


# --- Flags for converter/first_tranformation/section_transformation dicts

# A simple flag.
# The value is never considered; only the option string is returned.
FLAG = 0x01

# It's returned the option followed by the value.
VALUE = 0x02

# Call a function.
CALL = 0x04

# The corresponding option is returned if the value is 'true',
# the negation of the option is returned for 'false'.
BOOLEAN = 0x08

# Remove the attribute or section.
# A warning is issued only for sections.
REMOVE = 0x10

# If this section or attribute is present, a rule for this firewall
# can't be build.
FATAL = 0x20


def modify_dom(dom, convAttributes={}, convSections={},
                forEveryAttribute=None):
    """Modify a given DOM object. convAttributes is a dictionary used
    to modify the attributes, convSections is used to manage sections
    and forEveryAttribute is a function to be applied to every attribute.
    """
    o = dom.cloneNode(1)
    ni = NodeIterator(o, NodeFilter.SHOW_ELEMENT,
                        None, 0)
    cn = ni.nextNode()
    to_remove_sect = []
    while cn:
        # cn is the current node.
        # name is the name of the current node.
        name = cn.nodeName
        if name in convSections.keys():
            if convSections[name] & REMOVE:
                sl_print_warning('the "' + name + '" section' +
                                    ' will be ignored for this firewall')
                if cn not in to_remove_sect:
                    to_remove_sect.append(cn)
            elif convSections[name] & FATAL:
                sl_print_error('cannot build a rule with the "' +
                                    name + '" section')
                return None
        i = 0
        # for every attribute=value pair.
        while i < len(cn.attributes):
            n = cn.attributes.item(i).name
            v = cn.attributes.item(i).value
            sect = ''
            if convAttributes.get(name):
                sect = convAttributes[name]
            else:
                for s in convAttributes.keys():
                    if type(s) is type(()) and name in s:
                        sect = convAttributes[s]
                        break
            # Apply forEveryAttribute to every atttribute.
            if forEveryAttribute:
                sect = {n: (forEveryAttribute, CALL)}
            if sect:
                if sect.get(n):
                    to_set = ''
                    to_remove = 0
                    x = sect[n]
                    # n is the name of the attribute.
                    # v is the value of the attribute.
                    # a is the value of the option.
                    # f is the flag of the option.
                    a = x[0]
                    f = x[1]
                    if f & FLAG:
                        to_set = a
                    elif f & VALUE:
                        if a: a = a + ' '
                        to_set =  a + v
                    elif f & BOOLEAN:
                        if v == 'yes':
                            to_set = a
                        elif v == 'no':
                            to_set = '! ' + a
                    elif f & CALL:
                        try:
                            v2 = ''
                            if len(v) > 1 and v[0:2] == '! ':
                                v2 = '! '
                                v = v[2:]
                            to_set = a(v)
                            if v2: to_set = v2 + to_set
                        except RemoveOptionError:
                            to_remove = 1
                        except RemoveSectionError:
                            if cn not in to_remove_sect:
                                to_remove_sect.append(cn)
                            sl_print_warning('the "' + n + '" section' +
                                        ' will be ignored for this firewall')
                        except CreateRulesError:
                            sl_print_error('cannot build a rule with the "' +
                                            n + '" section or option')
                            return None
                    elif f & REMOVE:
                        to_remove = 1
                    elif f & FATAL:
                        sl_print_error('cannot build a rule with the "' +
                                        n + '" section')
                        return None
                    else:
                        sl_print_error('troubles in the convDict')
                    cn.setAttributeNS('', n, to_set)
                    if to_remove:
                        i -= 1
                        cn.removeAttributeNS('', n)
            i += 1
        cn = ni.nextNode()
    for s in to_remove_sect:
        if s.parentNode: s.parentNode.removeChild(s)
    return o


# --- Functions used to compare two DOM

def _compare_attrs(a1, a2):
    """Compare two lists of attributes."""
    if len(a1) != len(a2):
        return 0
    l1 = []
    k = 0
    while k < len(a1):
        item = a1.item(k)
        l1.append((item.name, item.value))
        k = k + 1
    l2 = []
    k = 0
    while k < len(a2):
        item = a2.item(k)
        l2.append((item.name, item.value))
        k = k + 1
    l1.sort()
    l2.sort()
    if l1 != l2:
        return 0
    return 1

def _compare_children(c1, c2):
    """Compare two list of children nodes."""
    l1 = []
    for i in c1:
        if i.nodeType == Node.ELEMENT_NODE:
            l1.append(i) 
    l2 = [] 
    for i in c2:
        if i.nodeType == Node.ELEMENT_NODE:
            l2.append(i) 
    if not l1 and not l2:
        return 1
    if len(l1) != len(l2):
        return 0
    for i in l1:
        for j in l2:
            if not compare_dom(i, j):
                break
        else:
            return 0
    return 1

def compare_dom(a, b):
    """Compare two DOMs.
    
    Given two DOMs, return 0 if they are equals.
    """
    if a.nodeType == Node.DOCUMENT_NODE:
        a = a.documentElement
    if b.nodeType == Node.DOCUMENT_NODE:
        b = b.documentElement
    if a.nodeName != b.nodeName:
        if not (a.nodeName in xml_commands and b.nodeName in xml_commands):
            return 1
    if not _compare_attrs(a.attributes, b.attributes):
        return 1
    if not _compare_children(a.childNodes, b.childNodes):
        return 1
    return 0


def flat_print_attrs(opt):
    """Return a string that is a space sparated list of the attribute values
    in the given XML."""
    strn = ''
    ni = NodeIterator(opt, NodeFilter.SHOW_ELEMENT, None, 0)
    cn = ni.nextNode()
    while cn:
        i = 0
        while i < len(cn.attributes):
            strn = strn + ' ' + cn.attributes.item(i).value
            i += 1
        cn = ni.nextNode()
    return strn


# --- Functions used to transform rule's attributes.

icmp_map = {
    'echo-reply':   (0,),
    'pong':         (0,),
    'destination-unreachable':      (3,),
    'network-unreachable':          (3, 0),
    'host-unreachable':             (3, 1),
    'protocol-unreachable':         (3, 2),
    'port-unreachable':             (3, 3),
    'fragmentation-needed':         (3, 4),
    'source-route-failed':          (3, 5),
    'network-unknown':              (3, 6),
    'host-unknown':                 (3, 7),
    'network-isolated':             (3, 8),
    'network-prohibited':           (3, 9),
    'host-prohibited':              (3, 10),
    'TOS-network-unreachable':      (3, 11),
    'TOS-host-unreachable':         (3, 12),
    'communication-prohibited':     (3, 13),
    'host-precedence-violation':    (3, 14),
    'precedence-cutoff':            (3, 15),
    'source-quench':    (4,),
    'redirect':             (5,),
    'network-redirect':     (5, 0),
    'host-redirect':        (5, 1),
    'TOS-network-redirect': (5, 2),
    'TOS-host-redirect':    (5, 3),
    'echo-request': (8,),
    'pong':         (8,),
    'router-advertisement': (9,),
    'router-solicitation':  (10,),
    'time-exceeded':                (11,),
    'ttl-exceeded':                 (11,),
    'ttl-zero-during-transit':      (11, 0),
    'ttl-zero-during-reassembly':   (11, 1),
    'parameter-problem':        (12,),
    'ip-header-bad':            (12, 0),
    'required-option-missing':  (12, 1),
    'timestamp-request':    (13,),
    'timestamp-reply':      (14,),
    'information-request':  (15,),
    'information-reply':    (16,),
    'address-mask-request': (17,),
    'address-mask-reply':   (18,)
}

def transf_icmp(v):
    """Return a numeric representation of icmp type/code."""
    r = v
    x = icmp_map.get(v)
    if x:
        r = str(x[0])
        if len(x) == 2: 
            r = r + '/' + str(x[1])
    return r

def transf_sort(v):
    """For options that need to be sorted.

    Given a string with comma separated items, these items are sorted;
    if portions of the string are sparated by slashes, the order of these
    groups is preserved.
    """
    r = []
    o = ''
    for s in v.split('/'):
        x = s.split(',')
        x.sort()
        r.append(x)
    for s in r:
        o2 = ''
        for t in s:
            if o2: o2 += ','
            o2 += t
        o += o2 + '/'
    return o[:-1]

def transf_ip(v):
    """Manage IP address.

    Return an IP in dotted notation with the right netmask.
    """
    host = v
    mask = '255.255.255.255'
    if v.find('/') != -1:
        host, mask = v.split('/')
    if host == '0':
        host = '0.0.0.0'
    # We have a host name.
    if re.search('[^0-9\.]', host):
        try:
            host = socket.gethostbyname(host)
        except socket.error:
            pass
    # Normalize host and mask.
    try:
        host = iplib.convert(host, notation=iplib.IP_DOT)
    except ValueError:
        pass
    try:
        mask = iplib.convert_nm(mask, notation=iplib.NM_DOT)
    except ValueError:
        pass
    if mask == '0.0.0.0':
        host = '0.0.0.0'
    hp = host.split('.')
    mp = mask.split('.')
    if len(hp) == len(mp) == 4:
        i = 0 
        while i < 4: 
            if mp[i] == '0':
                hp[i] = '0'
            i = i + 1
        host = hp[0] + '.' + hp[1] + '.' + hp[2] + '.' + hp[3]
    v = host + '/' + mask
    if v == '0.0.0.0/0.0.0.0':
        raise RemoveOptionError, 'remove the default value'
    return v

def transf_proto(v):
    """Return the protocol number.

    Given a protocol name, return the protocol number.
    """
    try:
        v = str(socket.getprotobyname(v))
    except socket.error:
        pass
    return v

def transf_port(v):
    """Return the port number.

    Given a port name, return the protocol number.
    """
    # XXX: ugly!
    plist = []
    if v.find(':') != -1:
        plist = v.split(':')
    else:
        plist = [v]
    index = 0
    # XXX: uh-uh... what if udp/tcp protocols names don't match?
    for i in plist:
        try:
            i = str(socket.getservbyname(i, 'tcp'))
        except socket.error:
            try:
                i = str(socket.getservbyname(i, 'udp'))
            except socket.error:
                pass
        plist[index] = i
        index = index + 1
    fin = plist[0]
    if len(plist) == 2:
        fin = fin + ':' + plist[1]
    if fin == '1:65535':
        raise RemoveOptionError, 'remove the default value'
    return fin

def transf_normalize(v):
    """Check for a space after an exclamation mark."""
    v = v.strip()
    if len(v) > 0 and v[0] == '!':
        return '! ' + v[1:].strip()
    return v

def transf_remove(v):
    """Simply remove an entry."""
    raise RemoveOptionError, 'remove this option'

def transf_fatal(v):
    """Raise an exception because a rule with this option/extension
    cannot be built with the currently used firewall."""
    raise CreateRulesError, 'Cannot insert a key in a rule for ' + \
                            'this firewall'

def transf_remove_section(d):
    """Raise an exception that force the modify_option function to
    completely discard the current section."""
    raise RemoveSectionError, 'cannot use this extension with this firewall'


class NodeFilterByElementName(NodeFilter):
    """Filter for NodeIterator.

    Select DOM nodes based on the name.
    """
    def __init__(self, n):
        if type(n) in (types.ListType, types.TupleType):
            self.names = n
        else:
            self.names = (n,)
    def acceptNode(self, node):
        if node.nodeName in self.names:
            return 1
        return 0


def set_dom_defaults(dom, defaults):
    """Set given default values in a DOM."""
    # TODO: convert 'defaults' from a dictionary to a DOM.
    for k, v in defaults.items():
        if type(k) in (types.ListType, types.TupleType) and \
                dom.nodeName in k:
            for k2, v2 in v.items():
                if not dom.getAttributeNS('', k2):
                    dom.setAttributeNS('', k2, v2)
        else:
            l = dom.getElementsByTagName(k)
            # XXX: add a switch that doesn't create a new element.
            if not l:
                o = dom.ownerDocument.createElement(k)
                dom.appendChild(o)
            else:
                o = l[0]
            for k2, v2 in v.items():
                if not o.getAttributeNS('', k2):
                    o.setAttributeNS('', k2, v2)
    return dom


# --- Functions used to import C modules

def get_C_module_name(n):
    """Return the expected name of the C module for the given firewall."""
    data = os.uname()
    opsys = data[0].lower()
    arch = data[4].lower()
    if len(arch) > 3 and arch[0] == 'i' and arch[2:4] == '86':
        arch = 'i386'
    return '_' + n + '_' + opsys + '_' + arch


def import_C_module(modname):
    """Import a C module; if the module doesn't exists, a fake
    module is returned."""
    modname = 'daxfi.firewalls.' + modname + '.' + get_C_module_name(modname)
    try: 
        m = __import__(modname, globals(), locals(),
                        [modname.split('.')[-1]])
    except ImportError:
        m = imp.new_module(modname)
    if not getattr(m, 'is_supported', None):
        def is_supported(*args, **kw): return 0
        setattr(m, 'is_supported', is_supported)
    if not getattr(m, 'list_rules', None):
        def list_rules(chain, *args, **kw): return []
        setattr(m, 'list_rules', list_rules)
    return m


