"""
ipchains package (daxfi.firewalls package).

Class used to manage ipchains rules.

  Copyright 2001, 2002 Davide Alberani <alberanid@libero.it>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

from daxfi.Rule import RuleBase
from daxfi._rulesutils import *
from daxfi._syslog import *

module = import_C_module('ipchains')

__version__ = '0.8'


def list_chains():
    return ('input', 'output', 'forward', 'nat')

def is_supported():
    return module.is_supported()

def list_rules(chain):
    if chain == 'nat':
        return module.list_rules('input', 1) + module.list_rules('forward', 1)
    return module.list_rules(chain)


def transf_ipchains_target(v):
    if v == 'drop':
        return '-j DENY'
    elif v == 'reject':
        return '-j REJECT'
    elif v == 'accept':
        return '-j ACCEPT'
    return v

def trasf_ipchains_nat(v):
    if v == 'snat':
        return 'masq'
    if v == 'dnat':
        return 'redirect'
    return v


# Sections to be modified for ipchains.
fw_section_transform_ipchains = {
    'limit':            REMOVE,
    'state':            FATAL
}

# - Sets of options for the ipchains firewall

fw_commands = {
    'append':           '-A',
    'delete':           '-D',
    'replace':          '-R',
    'insert':           '-I',
    'flush':            '-F',
    'set-policy':       '-P'
}

fw_rules_ipchains = {
    'source-ip':        ('--source', VALUE),
    'destination-ip':   ('--destination', VALUE),
    'interface':        ('--interface', VALUE),
    'fragment-only':    ('--fragment', BOOLEAN),
    'chain':            ('', VALUE)
}

fw_targets_ipchains = {
    'target':       (transf_ipchains_target, CALL),
    'reject-with':  ('', REMOVE)
}

fw_protocols_ipchains = {
    'protocol':         ('-p', VALUE),
    'source-port':      ('--source-port', VALUE),
    'destination-port': ('--destination-port', VALUE),
    'icmp-type':        ('', VALUE),
    'syn-only':         ('--syn', BOOLEAN),
    'tcp-flags':        ('', REMOVE)
}

fw_log_ipchains = {
    'log-priority':     ('', REMOVE),
    'log-facility':     ('', REMOVE)
}

fw_nat_ipchains = {
    'nat':          (trasf_ipchains_nat, CALL),
    'to-address':   ('', FATAL),
    'to-port':      ('', VALUE)
}

# The converter dictionary for ipchains.
fw_ipchains_converter = {
    xml_commands:   fw_rules_ipchains,
    'target':       fw_targets_ipchains,
    'protocol':     fw_protocols_ipchains,
    'nat':          fw_nat_ipchains
}


class Rule(RuleBase):
    """Methods and variables used to create the command string for ipchains."""
    converter = fw_ipchains_converter
    first_transformation = {}
    section_transformation = fw_section_transform_ipchains
    defaults = {}
    withLog = 1
    firewallName = 'ipchains'

    def _getRuleCommand(self):
        """Return a string used to run the command for this rule."""
        cmd = 'ipchains'

        opt = self.getModifiedDom()
        if not opt:
            sl_print_error('this rule is uninitialized')
            return ''

        action = self.getAction()
        exe = fw_commands.get(action)

        if not exe:
            sl_print_error('cannot get the action for this rule')
            return ''

        cmd = cmd + ' ' + exe

        has_nat = opt.getElementsByTagName('nat')
        nat = None
        nat_type = ''
        to_port = ''
        if has_nat:
            nat = has_nat[0]
            nat_type = nat.getAttributeNS('', 'nat')
            to_port = nat.getAttributeNS('', 'to-port')
            opt.removeChild(nat)

        has_target = opt.getElementsByTagName('target')
        target = ''
        if has_target and not nat:
            target = has_target[0].getAttributeNS('', 'target')
            has_target[0].removeAttributeNS('', 'target')
        chain = opt.getAttributeNS('', 'chain')
        if self.getAction() == 'flush':
            if chain == 'nat': chain = ''
            return cmd + ' ' + chain
        if self.getAction() == 'set-policy':
            if not (has_target and chain):
                sl_print_error('a set-policy command requires a target' + \
                                ' and a chain')
                return ''
            cmd = cmd + ' ' + chain
            cmd = cmd + ' ' + target[3:]
            return cmd

        if nat_type == 'redirect':
            cmd = cmd + ' input -j REDIRECT'
            if to_port:
                cmd = cmd + ' ' + to_port
        elif nat_type == 'masq':
            cmd = cmd + ' forward -j MASQ'
        if chain:
            cmd = cmd + ' ' + chain
            opt.removeAttributeNS('', 'chain')
        if self.getRuleNumber() != '-1' and \
            (self.getAction() == 'replace' or
                self.getAction() == 'insert' or self.getAction() == 'delete'):
            cmd = cmd + ' ' + self.getRuleNumber()
        if has_target:
            cmd = cmd + ' ' + target

        has_proto = opt.getElementsByTagName('protocol')
        itype = ''
        if has_proto:
            itype = has_proto[0].getAttributeNS('', 'icmp-type')
        if itype:
            if itype.find('/') == -1:
                cmd = cmd + ' --icmp-type ' + itype
            else:
                sp = has_proto[0].getAttributeNS('', 'source-port')
                dp = has_proto[0].getAttributeNS('', 'destination-port')
                if not sp: sp = '--source 0.0.0.0/0.0.0.0'
                if not dp: dp = '--destination 0.0.0.0/0.0.0.0'
                tc = itype.split('/')
                cmd = cmd + ' ' + sp + ' ' + tc[0] + \
                            ' ' + dp + ' ' + tc[1]
            has_proto[0].removeAttributeNS('', 'icmp-type')
        has_log = opt.getElementsByTagName('log')
        if has_log:
            cmd = cmd + ' -l'
            opt.removeChild(has_log[0])
        return cmd + flat_print_attrs(opt)


