"""
ipfilter package (daxfi.firewalls package).

Class used to manage ipfilter rules.

  Copyright 2002 Davide Alberani <alberanid@libero.it>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

from daxfi._rulesutils import *
from daxfi._syslog import *
from daxfi.Rule import RuleBase
from daxfi._exceptions import *

module = import_C_module('ipfilter')

__version__ = '0.8'


def list_chains():
    return ('input', 'output', 'nat')

def is_supported():
    return module.is_supported()

def list_rules(chain):
    if chain == 'nat':
        return module.list_nat()
    return module.list_rules(chain)


def transf_ipfilter_target(v):
    """Modify the target name."""
    v = v.lower()
    if v == 'drop':
        return 'block'
    if v == 'reject':
        return 'block return-icmp'
    if v == 'accept':
        return 'pass'
    return v

def transf_ipfilter_chain(v):
    """in and out chains for ipfilter."""
    if v == 'input':
        return 'in'
    if v == 'output':
        return 'out'
    return v

def transf_ipfilter_port(v):
    """Ipfilter use its own range syntax."""
    # XXX: quite messy!
    v = transf_port(v)
    if v.find(':') != -1:
        plist = v.split(':')
        a = plist[0]
        b = plist[1]
        return a + ' >< ' + b
    else:
        return '= ' + v

def transf_ipfilter_icmp(v):
    """Convert icmp names into icmp codes."""
    code = v.split('/')
    res = 'icmp-type ' + str(code[0])
    if len(code) > 1:
        res = res + ' code ' + str(code[1])
    return res

def transf_ipfilter_state(v):
    """Manage states for ipfilter."""
    if v == 'related':
        return 'keep state'
    raise CreateRulesError, 'unable to write a rule with this state'

def transf_ipfilter_syn(v):
    """ACK and SYN matches must be 'emulated' with 'flags'."""
    if v == 'yes':
        return 'flags S/SRA'
    elif v == 'no':
        return 'flags SA'
    return ''

def transf_ipfilter_logprio(v):
    """ipfilter uses the BSD names for logging priority."""
    if v == 'warning':
        v = 'warn'
    return v

def transf_ipfilter_tcpflags(v):
    """ipfilter uses it's own syntax for tcp flags."""
    out = ''
    if v.find(' ') != -1:
        vs = v.split(' ')
        v = vs[1] + '/' + vs[0]
    v = v.split('/')
    for e in v:
        x = e.split(',')
        for i in x:
            out = out + i[0]
        out = out + '/'
    return 'flags ' + out[:-1]


fw_section_transform_ipfilter = {
    'limit':            REMOVE
}

# - Sets of options for the ipfilter firewall

fw_commands = {
    'append':           '',
    'delete':           '',
    'insert':           '',
    'flush':            '',
    'set-policy':       ''
}

fw_rules_ipfilter = {
    'chain':            (transf_ipfilter_chain, CALL),
    'source-ip':        ('from', VALUE),
    'destination-ip':   ('to', VALUE),
    'interface':        ('on', VALUE),
    'fragment-only':    ('with frag', BOOLEAN),
    'chain':            (transf_ipfilter_chain, CALL),
}

fw_targets_ipfilter = {
    'target':       (transf_ipfilter_target, CALL),
    'reject-with':  ('', VALUE)
}

fw_protocols_ipfilter = {
    'protocol':         ('proto', VALUE),
    'source-port':      (transf_ipfilter_port, CALL),
    'destination-port': (transf_ipfilter_port, CALL),
    'icmp-type':        (transf_ipfilter_icmp, CALL),
    'tcp-flags':        (transf_ipfilter_tcpflags, CALL),
    'syn-only':         (transf_ipfilter_syn, CALL)
}

fw_log_ipfilter = {
    'log-priority':     (transf_ipfilter_logprio, CALL),
    'log-facility':     ('', VALUE)
}

fw_state_ipfilter = {
    'state':            (transf_ipfilter_state, CALL)
}

fw_nat_ipfilter = {
    'nat':          ('', VALUE),
    'to-address':   (transf_ip, CALL),
    'to-port':      ('', VALUE)
}

fw_ipfilter_converter = {
    xml_commands:   fw_rules_ipfilter,
    'target':       fw_targets_ipfilter,
    'protocol':     fw_protocols_ipfilter,
    'log':          fw_log_ipfilter,
    'nat':          fw_nat_ipfilter,
    'state':        fw_state_ipfilter
}


class Rule(RuleBase):
    """Methods and variables used to create the command string for ipfilter."""
    converter = fw_ipfilter_converter
    first_transformation = {}
    section_transformation = fw_section_transform_ipfilter
    defaults = {}
    firewallName = 'ipfilter'
    withLog = 1

    def __str__(self):
        """Print this rule."""
        cmd = self._buildRule()
        if not cmd: return ''
        if cmd[0] == '&':
            return '# Execute the command: ' + cmd[1:]
        elif cmd[0] == '%':
            return '# Remove the rule: "' + cmd[1:] + '"'
        elif cmd[0] == '^':
            return '# Remove the NAT rule: "' + cmd[1:] + '"'
        if cmd[0] == '!':
            return cmd[1:]
        else:
            return cmd

    def _getRuleCommand(self):
        """Return a string that is the command to run ipfilter rules."""
        cmd = self._buildRule()
        if not cmd:
            return ''
        if cmd[0] == '&':
            return cmd[1:]
        elif cmd[0] == '!':
            return 'echo "' + cmd[1:] + '" | ipnat -f -'
        elif cmd[0] == '%':
            return 'echo "' + cmd[1:] + '" | ipf -rf -'
        elif cmd[0] == '^':
            return 'echo "' + cmd[1:] + '" | ipnat -rf -'
        else:
            return 'echo "' + cmd + '" | ipf -f -'
        
    def _buildRule(self):
        """Internal function that builds the command string for
        this rule.  Commands to be executed directly are prepended
        with a &; other lines are commands from a configuration file
        for ipfilter."""
        cmd = ''

        action = self.getAction()
        exe = fw_commands.get(action)

        if exe is None:
            sl_print_error('cannot get the action for this rule')
            return ''

        opt = self.getModifiedDom()
        if not opt:
            return '' 

        chain = opt.getAttributeNS('', 'chain')
        if chain == 'forward':
            sl_print_warning('cannot use the "forward" chain with ipfilter')
            return ''

        has_nat = opt.getElementsByTagName('nat')
        nat = None
        nat_type = ''
        to_port = ''
        to_address = ''
        if has_nat:
            nat = has_nat[0]
            nat_type = nat.getAttributeNS('', 'nat')
            to_port = nat.getAttributeNS('', 'to-port')
            to_address = nat.getAttributeNS('', 'to-address')
            opt.removeChild(nat)
            if to_port == '0': to_port = ''

        if action == 'flush':
            if chain == 'nat': return '&ipnat -C'
            return '&ipf -F' + chain[0]

        if nat:
            if nat_type in ('redirect', 'dnat'):
                cmd = 'rdr'
            elif nat_type in ('masq', 'snat'):
                cmd = 'map'

        target = opt.getElementsByTagName('target')
        tg = ''
        if target and not nat:
            tg = target[0].getAttributeNS('', 'target')
            if tg:
                target[0].removeAttributeNS('', 'target')
            rw = target[0].getAttributeNS('', 'reject-with')
            if rw:
                tg = 'block return-icmp'
                spl = rw.split('/')
                rw = spl[0]
                rc = ''
                if len(spl) > 1:
                    rc = spl[1]
                if rw == '3' and rc:
                    rw = rc
                tg = tg + '(' + rw + ')'

        if action == 'set-policy':
            return cmd + tg + ' ' + chain + ' all'
 
        if action == 'insert':
            nr = self.getRuleNumber()
            if nr and nr != '-1':
                cmd = cmd + ' @' + nr + ' '
        if exe:
            cmd = cmd + ' ' + exe
        if tg:
            cmd = cmd + tg

        chain = opt.getAttributeNS('', 'chain')
        st = ''
        state = opt.getElementsByTagName('state')
        if state:
            st = state[0].getAttributeNS('', 'state')
            if st:
                state[0].removeAttributeNS('', 'state')
        if chain:
            if state:
                if st == 'keep state':
                    if chain == 'in':
                        chain = 'out'
                    elif chain == 'out':
                        chain = 'in'
            if not nat:
                cmd = cmd + ' ' + chain
            opt.removeAttributeNS('', 'chain')

        has_log = opt.getElementsByTagName('log')
        if has_log:
            cmd = cmd + ' log'
            fac = has_log[0].getAttributeNS('', 'log-facility')
            lev = has_log[0].getAttributeNS('', 'log-priority')
            if fac:
                if not lev:
                    lev = 'warn'
                lev = fac + '.' + lev
            if lev:
                cmd = cmd + ' level ' + lev
            opt.removeChild(has_log[0])

        # Always use the quick option.
        if not nat:
            cmd = cmd + ' quick'

        iface = opt.getAttributeNS('', 'interface')
        if iface:
            opt.removeAttributeNS('', 'interface')
            if nat: iface = iface[3:]
            cmd = cmd + ' ' + iface

        has_proto = opt.getElementsByTagName('protocol')
        proto = None
        if has_proto:
            proto = has_proto[0].getAttributeNS('', 'protocol')
        if proto:
            if not nat: cmd = cmd + ' ' + proto
            has_proto[0].removeAttributeNS('', 'protocol')

        sip = opt.getAttributeNS('', 'source-ip')
        if sip:
            opt.removeAttributeNS('', 'source-ip')
        else:
            sip = 'from 0.0.0.0/0.0.0.0'
        #if nat_type in ('redirect', 'dnat'): sip = sip[5:]
        cmd = cmd + ' ' + sip
        if has_proto:
            sp = has_proto[0].getAttributeNS('', 'source-port')
            if sp:
                if len(sp) > 1 and sp[0] == '!':
                    if sp[2] == '=':
                        sp = '!= ' + sp[3:]
                    else:
                        sp = sp[2:]
                        sp = sp.replace('><', '<>')
                #if nat_type in ('redirect', 'dnat'): sp = sp.replace('=', '')
                cmd = cmd + ' port ' + sp
                has_proto[0].removeAttributeNS('', 'source-port')
        dip = opt.getAttributeNS('', 'destination-ip')
        if dip:
            opt.removeAttributeNS('', 'destination-ip')
        else:
            dip = 'to 0.0.0.0/0.0.0.0'
        #if nat_type not in ('redirect', 'dnat'):
        cmd = cmd + ' ' + dip
        if has_proto:
            dp = has_proto[0].getAttributeNS('', 'destination-port')
            if dp:
                if len(dp) > 1 and dp[0] == '!':
                    if dp[2] == '=':
                        dp = '!= ' + dp[3:]
                    else:
                        dp = dp[2:]
                        dp = dp.replace('><', '<>')
                #if nat_type in ('redirect', 'dnat'):
                #    dp = dp.replace(' =', '')
                cmd = cmd + ' port ' + dp
                has_proto[0].removeAttributeNS('', 'destination-port')
            tcpflags = has_proto[0].getAttributeNS('', 'tcp-flags')
            if tcpflags:
                if not nat: cmd = cmd + ' ' + tcpflags
                has_proto[0].removeAttributeNS('', 'tcp-flags')
            syn = has_proto[0].getAttributeNS('', 'syn-only')
            if syn:
                if not nat: cmd = cmd + ' ' + syn
                has_proto[0].removeAttributeNS('', 'syn-only')
        frag = opt.getAttributeNS('', 'fragment-only')
        if frag:
            if not nat: cmd = cmd + ' ' + frag
            opt.removeAttributeNS('', 'fragment-only')
        if has_proto:
            icmptype = has_proto[0].getAttributeNS('', 'icmp-type')
            if icmptype:
                if not nat: cmd = cmd + ' ' + icmptype
                has_proto[0].removeAttributeNS('', 'icmp-type')

        if st and not nat:
            cmd = cmd + ' ' + st

        if nat:
            cmd = cmd + ' ->'
            if nat_type == 'masq':
                to_address = '0/32'
            elif nat_type == 'redirect':
                to_address = '127.0.0.1'
            if not to_address: to_address = '0.0.0.0/0.0.0.0'
            cmd = cmd + ' ' + to_address
            if proto: proto = proto[6:]
            if to_port:
                cmd = cmd + ' port'
                if nat_type in ('snat', 'masq'):
                    cmd = cmd + 'map' # the 'portmap' keyword
                    if proto == '17': cmd = cmd + ' udp'
                    else: cmd = cmd + ' tcp'
                    cmd = cmd + ' ' + to_port
                else:
                    cmd = cmd + ' ' + to_port
                    if proto == '17': cmd = cmd + ' udp'
                    elif proto == '6': cmd = cmd + ' tcp'

        if action == 'delete':
            if nat: return '^' + cmd
            rn = self.getRuleNumber()
            if rn and rn != '-1':
                sl_print_error('cannot delete a rule via rule-number')
                return ''
            return '%' + cmd

        if nat:
            return '!' + cmd
        return cmd

