"""
ipfwadm package (daxfi.firewalls package).

Class used to manage ipfwadm rules.

  Copyright 2001, 2002 Davide Alberani <alberanid@libero.it>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

from daxfi.Rule import RuleBase
from daxfi._rulesutils import *
from daxfi._syslog import *

module = import_C_module('ipfwadm')

__version__ = '0.8'


def list_chains():
    return ('input', 'output', 'forward', 'nat')

def is_supported():
    return module.is_supported()

def list_rules(chain):
    if chain == 'nat':
        return module.list_rules('forward', 1) + module.list_rules('input', 1)
    return module.list_rules(chain)

def transf_ipfwadm_proto(v):
    """Ipfwadm wants a protocol name, instead of a protocol number."""
    if v == '0' or v == 'all':
        return '-P all'
    elif v == '1' or v == 'icmp':
        return '-P icmp'
    elif v == '6' or v == 'tcp':
        return '-P tcp'
    elif v == '17' or v == 'udp':
        return '-P udp'
    return '-P' + v

def trasf_ipfwadm_icmptype(v):
    """Ipfwadm need an icmp type expressed in numeric format."""
    i = v.find('/')
    if i != -1:
        v = v[:i]
    return v

def transf_ipfwadm_target(v):
    if v == 'drop':
        return 'deny'
    return v

def trasf_ipfwadm_nat(v):
    if v == 'snat':
        return 'masq'
    if v == 'dnat':
        return 'redirect'
    return v


# Sections to be modified for ipfwadm.
fw_section_transform_ipfwadm = {
    'limit':            REMOVE,
    'state':            FATAL
}

# - Sets of options for the ipfwadm firewall

fw_commands = {
    'append':           '-a',
    'delete':           '-d',
    'insert':           '-i',
    'flush':            '-f',
    'set-policy':       '-p'
}

fw_rules_ipfwadm = {
    'source-ip':        ('-S', VALUE),
    'destination-ip':   ('-D', VALUE),
    'interface':        ('-W', VALUE),
    'chain':            ('', VALUE),
    'fragment-only':    ('', FATAL)
}

fw_target_ipfwadm = {
    'target':           (transf_ipfwadm_target, CALL),
    'reject-with':      ('', REMOVE)
}

fw_protocols_ipfwadm = {
    'protocol':         (transf_ipfwadm_proto, CALL),
    'source-port':      ('', VALUE),
    'destination-port': ('', VALUE),
    'icmp-type':        (trasf_ipfwadm_icmptype, CALL),
    'tcp-flags':        ('', REMOVE),
    'syn-only':         ('-y', BOOLEAN)
}

fw_log_ipfwadm = {
    'log-priority':     ('', REMOVE),
    'log-facility':     ('', REMOVE)
}

fw_nat_ipfwadm = {
    'nat':          (trasf_ipfwadm_nat, CALL),
    'to-address':   ('', FATAL),
    'to-port':      ('-r', VALUE)
}

# The converter dictionary for ipfwadm.
fw_ipfwadm_converter = {
    xml_commands:   fw_rules_ipfwadm,
    'protocol':     fw_protocols_ipfwadm,
    'target':       fw_target_ipfwadm,
    'log':          fw_log_ipfwadm,
    'nat':          fw_nat_ipfwadm
}


class Rule(RuleBase):
    """Methods and variables used to create the command string for ipfwadm."""
    converter = fw_ipfwadm_converter
    first_transformation = {}
    section_transformation = fw_section_transform_ipfwadm
    defaults = {}
    withLog = 1

    def _getRuleCommand(self):
        """Return a string used to run the command for this rule."""
        cmd = 'ipfwadm'

        action = self.getAction()
        if not action:
            sl_print_error('cannot get the action for this rule')
            return ''

        exe = fw_commands.get(action)
        if exe == None:
            sl_print_error('unsupported action for ipfwadm: ' + action)
            return ''
        cmd = cmd + ' ' + exe
        opt = self.getModifiedDom()

        if not opt:
            sl_print_error('empty rule')
            return ''

        rn = self.getRuleNumber()
        if action == 'delete' and rn and rn != '-1':
            sl_print_error('cannot delete a rule via rule-number')
            return ''

        if action == 'insert' and rn and rn != '-1':
            sl_print_warning('cannot insert a rule in a arbitrary ' + \
                                'position, insert at the top')

        has_nat = opt.getElementsByTagName('nat')
        nat = None
        nat_type = ''
        to_port = ''
        if has_nat:
            nat = has_nat[0]
            nat_type = nat.getAttributeNS('', 'nat')
            to_port = nat.getAttributeNS('', 'to-port')
            opt.removeChild(nat)

        has_target = opt.getElementsByTagName('target')
        if has_target and not nat:
            target = has_target[0].getAttributeNS('', 'target')
            has_target[0].removeAttributeNS('', 'target')
            cmd = cmd + ' ' + target

        if nat_type == 'redirect':
            cmd = cmd + ' accept -I'
        elif nat_type == 'masq':
            cmd = cmd + ' masquerade -F'
        if to_port:
            cmd = cmd + ' ' + to_port

        chn = opt.getAttributeNS('', 'chain')
        if chn and not nat:
            opt.removeAttributeNS('', 'chain')
            if chn == 'input':
                cmd = cmd + ' -I'
            elif chn == 'output':
                cmd = cmd + ' -O'
            elif chn == 'forward':
                cmd = cmd + ' -F'

        if action == 'flush' and chn == 'nat':
            sl_print_error('cannot flush NAT tables for ipfwadm; instead ' + \
                            'flush input, forward and output chains separately')
            return ''

        has_proto = opt.getElementsByTagName('protocol')
        if has_proto:
            pt = has_proto[0].getAttributeNS('', 'protocol')
            if pt == '-P all':
                has_proto[0].removeAttributeNS('', 'protocol')

            sp = has_proto[0].getAttributeNS('', 'source-port')
            if not sp:
                sp = has_proto[0].getAttributeNS('', 'icmp-type')
                if sp:
                    has_proto[0].removeAttributeNS('', 'icmp-type')
            else:
                has_proto[0].removeAttributeNS('', 'source-port')
            if sp:
                sa = opt.getAttributeNS('', 'source-ip')
                if sa:
                    opt.removeAttributeNS('', 'source-ip')
                if not sa: sa = '-S 0.0.0.0/0.0.0.0'
                cmd = cmd + ' ' + sa + ' ' + sp

            dp = has_proto[0].getAttributeNS('', 'destination-port')
            if dp:
                has_proto[0].removeAttributeNS('', 'destination-port')
                da = opt.getAttributeNS('', 'destination-ip')
                if da:
                    opt.removeAttributeNS('', 'destination-ip')
                if not da: da = '-D 0.0.0.0/0.0.0.0'
                cmd = cmd + ' ' + da + ' ' + dp
            syn = has_proto[0].getAttributeNS('', 'syn-only')
            if syn == '! -y':
                has_proto[0].removeAttributeNS('', 'syn-only')
                cmd = cmd + ' -k'

        has_log = opt.getElementsByTagName('log')
        if has_log:
            has_log[0].parentNode.removeChild(has_log[0])
            cmd = cmd + ' -o'
        cmd = cmd + ' ' + flat_print_attrs(opt)
        if cmd.find('!') != -1:
            sl_print_error('ipfwadm does not support parameter negation')
            return ''
        return cmd


