"""
iptables package (daxfi.firewalls package).

Class used to manage iptables rules.

  Copyright 2001, 2002 Davide Alberani <alberanid@libero.it>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

from daxfi.Rule import RuleBase
from daxfi._rulesutils import *
from daxfi._syslog import *

module = import_C_module('iptables')

__version__ = '0.8'


def list_chains():
    return ('input', 'output', 'forward', 'nat')

def list_rules(chain):
    if chain == 'nat':
        return module.list_rules('POSTROUTING', 1) + \
                module.list_rules('PREROUTING', 1) + \
                module.list_rules('OUTPUT', 1)
    return module.list_rules(chain)

def is_supported():
    return module.is_supported()


def transf_iptables_chain(v):
    if v in ('input', 'output', 'forward'):
        return v.upper()
    return v

def transf_iptables_target(v):
    if v in ('drop', 'reject', 'accept'):
        return '-j ' + v.upper()
    return '-j ' + v

def transf_iptables_rate(v):
    """Transform the rate of the iptables' limit extension."""
    ret = v
    index = v.find('/')
    if index != -1:
        val = int(v[:index])
        if val < 0:
            return '--rate ' + ret
        unit = v[index+1:]
        mul = 1
        if unit == 's' or unit == 'second':
            mul = 1
        elif unit == 'm' or unit == 'minute':
            mul = 60
        elif unit == 'h' or unit == 'hour':
            mul = 60*60
        elif unit == 'd' or unit == 'day':
            mul = 24*60*60
        ret = str(mul / val)
    return '--rate ' + ret

def transf_iptables_state(v):
    if v == 'new':
        return '--state INVALID,NEW'
    elif v == 'related':
        return '--state ESTABLISHED,RELATED'
    return v

def transf_iptables_tcpflags(v):
    """Invert value/mask for iptables' tcp-flags."""
    r = v.upper()
    if v.find('/') != -1:
        l = v.split('/')
        l.reverse()
        r = l[0] + '/' + l[1]
    return '--tcp-flags ' + r

def transf_iptables_rejectwith(v):
    if v == '3/0':
        return '--reject-with icmp-net-unreachable'
    if v == '3/1':
        return '--reject-with icmp-host-unreachable'
    if v == '3/2':
        return '--reject-with icmp-proto-unreachable'
    if v == '3/3':
        return '--reject-with icmp-port-unreachable'
    if v == '3/9':
        return '--reject-with icmp-net-prohibited'
    if v == '3/10':
        return '--reject-with icmp-host-prohibited'
    return '--reject-with ' + v

def transf_iptables_nattype(v):
    if v == 'redirect':
        return '-j REDIRECT'
    if v == 'masq':
        return '-j MASQUERADE'
    if v == 'snat':
        return '-j SNAT'
    if v == 'dnat':
        return '-j DNAT'
    return v


# - Sets of options for the iptables/netfilter firewall

fw_commands = {
    'append':       '-A',
    'delete':       '-D',
    'replace':      '-R',
    'insert':       '-I',
    'flush':        '-F',
    'set-policy':   '-P'
}

fw_rules_iptables = {
    'source-ip':        ('--source', VALUE),
    'destination-ip':   ('--destination', VALUE),
    'interface':        ('--in-interface', VALUE),
    'fragment-only':    ('--fragment', BOOLEAN),
    'chain':            (transf_iptables_chain, CALL)
}

fw_targets_iptables = {
    'target':       (transf_iptables_target, CALL),
    'reject-with':  (transf_iptables_rejectwith, CALL)
}

fw_protocols_iptables = {
    'protocol':         ('-p', VALUE),
    'source-port':      ('--source-port', VALUE),
    'destination-port': ('--destination-port', VALUE),
    'tcp-flags':        (transf_iptables_tcpflags, CALL),
    'icmp-type':        ('--icmp-type', VALUE),
    'syn-only':         ('--syn', BOOLEAN)
}

fw_log_iptables = {
    'log-priority':     ('--log-level', VALUE),
    'log-facility':     ('', REMOVE)
}

fw_limits_iptables = {
    'rate':      (transf_iptables_rate, CALL),
    'burst':     ('--limit-burst', VALUE)
}

fw_state_iptables = {
    'state':            (transf_iptables_state, CALL)
}

fw_nat_iptables = {
    'nat':          (transf_iptables_nattype, CALL),
    'to-address':   ('', VALUE),
    'to-port':      ('', VALUE)
}

# The converter dictionary for iptables.
fw_iptables_converter = {
    xml_commands:   fw_rules_iptables,
    'target':       fw_targets_iptables, 
    'protocol':     fw_protocols_iptables,
    'log':          fw_log_iptables,
    'limit':        fw_limits_iptables,
    'state':        fw_state_iptables,
    'nat':          fw_nat_iptables
}


class Rule(RuleBase):
    """Methods and variables used to create the command string for iptables."""
    converter = fw_iptables_converter
    first_transformation = {}
    section_transformation = {}
    defaults = {}
    firewallName = 'iptables'
    withLog = 0

    def _getRuleCommand(self):
        """Return a string used to run the command for this rule."""
        cmd = 'iptables'
        
        opt = self.getModifiedDom()
        if not opt:
            sl_print_error('this rule is uninitialized')
            return ''

        action = self.getAction()
        if not action:
            sl_print_error('cannot get the action for this rule')
            return ''

        exe = fw_commands[action]
        cmd = cmd + ' ' + exe
        has_target = opt.getElementsByTagName('target')
        chain = opt.getAttributeNS('', 'chain')
        has_proto = opt.getElementsByTagName('protocol')

        if action == 'set-policy':
            if not (has_target and chain):
                sl_print_error('a set-policy command requires a target' + \
                                ' and a chain')
                return ''
            t = has_target[0].getAttributeNS('', 'target')
            cmd = cmd + ' ' + chain
            cmd = cmd + ' ' + t[3:]
            return cmd

        if action == 'flush':
            if chain == 'nat': chain = ''
            return cmd + ' ' + chain

        has_nat = opt.getElementsByTagName('nat')
        nat = None
        nat_type = ''
        to_port = ''
        to_address = ''
        if has_nat:
            nat = has_nat[0]
            if chain:
                opt.removeAttributeNS('', 'chain')
            nat_type = nat.getAttributeNS('', 'nat')
            to_port = nat.getAttributeNS('', 'to-port')
            to_address = nat.getAttributeNS('', 'to-address')
            opt.removeChild(nat)
            if to_port == '0': to_port = ''
            to_port = to_port.replace(':', '-')
            to_address = to_address.replace(':', '-')

        if nat_type == '-j REDIRECT':
            chain = 'OUTPUT'
        elif nat_type == '-j MASQUERADE':
            chain = 'POSTROUTING'
        elif nat_type == '-j DNAT':
            chain = 'OUTPUT'
        elif nat_type == '-j SNAT':
            chain = 'POSTROUTING'

        if chain:
            cmd = cmd + ' ' + chain
            if not nat:
                opt.removeAttributeNS('', 'chain')

        if self.getRuleNumber() != '-1' and \
                (action in ('replace', 'insert', 'delete')):
            cmd = cmd + ' ' + self.getRuleNumber()

        if has_proto:
            p = has_proto[0].getAttributeNS('', 'protocol')
            if p:
                cmd = cmd + ' ' + p
                has_proto[0].removeAttributeNS('', 'protocol')
        elif to_port:
            cmd = cmd + ' -p 6'
 
        if nat:
            cmd = cmd + ' -t nat ' + nat_type
        if nat_type == '-j REDIRECT' or nat_type == '-j MASQUERADE':
            if to_port:
                cmd = cmd + ' --to-ports ' + to_port
        elif nat_type == '-j DNAT':
            if to_address:
                cmd = cmd + ' --to-destination ' + to_address
            if to_port:
                cmd = cmd + ':' + to_port
        elif nat_type == '-j SNAT':
            if to_address:
                cmd = cmd + ' --to-source ' + to_address
            if to_port:
                cmd = cmd + ':' + to_port

        if chain in ('OUTPUT', 'POSTROUTING'):
            i = opt.getAttributeNS('', 'interface')
            if i:
                opt.setAttributeNS('', 'interface', '--out-interface ' + \
                                   i.split()[1])
        if opt.getElementsByTagName('state'):
            cmd = cmd + ' -m state'

        has_limit = opt.getElementsByTagName('limit')
        if has_limit:
            cmd = cmd + ' -m limit'
            rate = has_limit[0].getAttributeNS('', 'rate')
            if rate:
                try:
                    rate = int(rate[8:])
                except ValueError:
                    rate = 1
            if rate != 0:
                rate = 10000 / rate
                has_limit[0].setAttributeNS('', 'rate', '--limit ' + str(rate))

        has_log = opt.getElementsByTagName('log')
        if has_log:
            cmd = cmd + ' -j LOG'
            if has_target:
                has_target[0].removeAttributeNS('', 'target')

        if has_target:
            tg = has_target[0].getAttributeNS('', 'target')
            cmd = cmd + ' ' + tg
            if tg:
                has_target[0].removeAttributeNS('', 'target')

        return cmd + flat_print_attrs(opt)


