"""
iplib module (daxfi package).

The representations of an IPv4 address and a netmask.
You can use this module to convert amongst many different IP notations.

This module is part of the IPConv program, that you can find here:
http://digilander.iol.it/alberanid/soft/

  Copyright 2001, 2002 Davide Alberani <alberanid@libero.it>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

import re, string, types

__version__ = '0.8'

# So that we can work even with Python 2.0 and unicode chars.
try:
    _StringTypes = (types.StringType, types.UnicodeType)
except AttributeError:
    _StringTypes = (types.StringType,)


# --- IpAddress

# Notation types.
IP_DOT = 1
IP_HEX = 2
IP_BIN = 3
IP_OCT = 4
IP_DEC = 5
IP_UNKNOWN = 0

# Some regexp.
_hex_match = re.compile('^(0x)?[0-9a-f]{1,8}$', re.IGNORECASE)
_oct_match = re.compile('^0[0-7]{1,12}$')
_bin_match = re.compile('^[01]{32}$')


def is_dot(ip):
    """Return true if the address is in dotted decimal notation."""
    if type(ip) not in _StringTypes:
        raise ValueError, 'is_dot: the ip must be a string'
    octets = string.split(ip, '.')
    if len(octets) != 4:
        return 0
    for i in octets:
        try:
            val = long(i)
        except ValueError:
            return 0
        if val < 0 or val > 255:
            return 0
    return 1

def is_hex(ip):
    """Return true if the address is in hexadecimal notation."""
    if type(ip) not in _StringTypes:
        raise ValueError, 'is_hex: the ip must be a string'
    if not _hex_match.match(ip):
        return 0
    try:
        dec = string.atol(ip, 16)
    except ValueError:
        return 0
    if dec < 0 or dec > 0xFFFFFFFFL:
        return 0
    return 1

def is_bin(ip):
    """Return true if the address is in binary notation."""
    if type(ip) not in _StringTypes:
        raise ValueError, 'is_bin: the ip must be a string'
    if not _bin_match.match(ip):
        return 0
    return 1

def is_oct(ip):
    """Return true if the address is in octal notation."""
    if type(ip) not in _StringTypes:
        raise ValueError, 'is_oct: the ip must be a string'
    if not _oct_match.match(ip):
        return 0
    try:
        dec = string.atol(ip, 8)
    except ValueError:
        return 0
    if dec < 0 or dec > 037777777777L:
        return 0
    return 1

def is_dec(ip):
    """Return true if the address is in decimal notation."""
    if type(ip) not in _StringTypes:
        raise ValueError, 'is_dec: the ip must be a string'
    try:
        dec = long(ip)
    except ValueError:
        return 0
    if dec < 0 or dec > 4294967295L:
        return 0
    return 1


def detect(ip):
    """Return the kind of an Internet address.

    Given an ip, returns the kind of the notation (one of
    IP_DOT, IP_HEX, IP_BIN, IP_DEC, IP_UNKNOWN).
    """
    if is_dot(ip):
        return IP_DOT
    elif is_bin(ip):
        return IP_BIN
    elif is_oct(ip):
        return IP_OCT
    elif is_hex(ip):
        return IP_HEX
    elif is_dec(ip):
        return IP_DEC
    return IP_UNKNOWN

def p_detect(ip):
    """Return the kind of an Internet address (string).

    Given an ip, returns a string that represents the kind of notation.
    """
    kind = detect(ip)
    if kind == IP_DOT:
        return 'dotted decimal'
    elif kind == IP_HEX:
        return 'hexadecimal'
    elif kind == IP_BIN:
        return 'binary'
    elif kind == IP_OCT:
        return 'octal'
    elif kind == IP_DEC:
        return 'decimal'
    return 'unknown'


def _dot_to_dec(ip):
    """Dotted decimal notation to decimal conversion."""
    if not is_dot(ip):
        raise ValueError, '_dot_to_dec: ip not valid'
    octets = string.split(ip, '.')
    dec = 0L
    dec = dec | (long(octets[0]) << 24)
    dec = dec | (long(octets[1]) << 16)
    dec = dec | (long(octets[2]) << 8)
    dec = dec | (long(octets[3]))
    return repr(dec)[:-1]

def _dec_to_dot(ip):
    """Decimal to dotted decimal notation convesion."""
    if not is_dec(ip):
        raise ValueError, '_dec_to_dot: ip not valid'
    ip = long(ip)
    first = int((ip >> 24) & 255)
    second =  int((ip >> 16) & 255)
    third = int((ip >> 8) & 255)
    fourth = int(ip & 255)
    return repr(first) + '.' + repr(second) + '.' + \
            repr(third) + '.' + repr(fourth)


def _hex_to_dec(ip):
    """Hexadecimal to decimal conversion."""
    if type(ip) is types.IntType:
        ip = hex(ip)
    elif type(ip) is types.LongType:
        ip = hex(ip)[:-1]
    if not is_hex(ip):
        raise ValueError, '_hex_to_dec: ip not valid'
    dec = string.atol(ip, 16)
    return repr(dec)[:-1]

def _dec_to_hex(ip):
    """Decimal to hexadecimal convesion."""
    if not is_dec(ip):
        raise ValueError, '_dec_to_hex: ip not valid'
    ip = long(ip)
    return hex(ip)[:-1]


def _oct_to_dec(ip):
    """Octal to decimal conversion."""
    if type(ip) is types.IntType:
        ip = oct(ip)
    elif type(ip) is types.LongType:
        ip = oct(ip)[:-1]
    if not is_oct(ip):
        raise ValueError, '_oct_to_dec: ip not valid'
    dec = string.atol(ip, 8)
    return repr(dec)[:-1]

def _dec_to_oct(ip):
    """Decimal to octal convesion."""
    if not is_dec(ip):
        raise ValueError, '_dec_to_oct: ip not valid'
    ip = long(ip)
    return oct(ip)[:-1]


def _bin_to_dec(ip):
    """Binary to decimal convesion."""
    if not is_bin(ip):
        raise ValueError, '_bin_to_dec: ip not valid'
    dec = 0L
    for i in range(0, 32):
        dec = dec | (long(ip[i]) << (31 - i))
    return repr(dec)[:-1]

def _dec_to_bin(ip):
    """Decimal to binary conversion."""
    if not is_dec(ip):
        raise ValueError, '_dec_to_bin: ip not valid'
    ip = long(ip)
    bin = ''
    for i in range(0, 32):
        if ip & 1:
            bin = '1' + bin
        else:
            bin = '0' + bin
        ip = ip >> 1
    return bin

def _dec_to_dec(ip):
    """Decimal to decimal conversion."""
    if type(ip) is types.IntType:
        ip = repr(ip)
    elif type(ip) is types.LongType:
        ip = repr(ip)[:-1]
    if not is_dec(ip):
        raise ValueError, '_dec_to_dec: ip not valid'
    return ip


def convert(ip, notation=IP_DOT, iformat=IP_UNKNOWN):
    """Convert among ip notations.

    Given an ip address, this functions returns the ip in another
    notation.

    *ip* -- The ip address.

    *notation* -- The format of the output.

    *iformat* -- Force the input to be considered of the given type
                    (otherwise it's autodetected).
    """
    kind = iformat
    if type(ip) in (types.IntType, types.LongType):
        if iformat not in (IP_DEC, IP_HEX, IP_OCT):
            raise ValueError, 'convert: if the ip is given in a numeric' + \
                            ' format, you must specify it with the "iformat"'+\
                            ' parameter'
    if kind == IP_UNKNOWN:
        kind = detect(ip)
    dec = 0L
    if kind == IP_DOT:
        dec = _dot_to_dec(ip)
    elif kind == IP_HEX:
        dec = _hex_to_dec(ip)
    elif kind == IP_BIN:
        dec = _bin_to_dec(ip)
    elif kind == IP_OCT:
        dec = _oct_to_dec(ip)
    elif kind == IP_DEC:
        dec = _dec_to_dec(ip)
    else:
        raise ValueError, 'convert: unknown ip format: "' + str(kind) + '"'
    if notation == IP_DOT:
        return _dec_to_dot(dec)
    elif notation == IP_HEX:
        return _dec_to_hex(dec)
    elif notation == IP_BIN:
        return _dec_to_bin(dec)
    elif notation == IP_OCT:
        return _dec_to_oct(dec)
    elif notation == IP_DEC:
        return dec
    else:
        raise ValueError, 'convert: unknown notation: "' + str(notation) + '"'


class IpAddress:
    """An IPv4 Internet address.

    This class represents an IPv4 Internet address.
    """
    def __init__(self, ip=''):
        if ip:
            self.set(ip)
        else:
            self.ip = ''

    def set(self, ip):
        """Set the IP address."""
        self.ip = convert(ip, detect(ip))

    def get(self):
        """Return the address."""
        return self.ip

    def get_dot(self):
        """Return the dotted decimal notation of the address."""
        return convert(self.ip, notation=IP_DOT)

    def get_hex(self):
        """Return the hexadecimal notation of the address."""
        return convert(self.ip, notation=IP_HEX)

    def get_bin(self):
        """Return the binary notation of the address."""
        return convert(self.ip, notation=IP_BIN)

    def get_dec(self):
        """Return the decimal notation of the address."""
        return convert(self.ip, notation=IP_DEC)

    def get_oct(self):
        """Return the octal notation of the address."""
        return convert(self.ip, notation=IP_OCT)


# --- IpNetMask

# XXX: 255.255.255.254 is not a valid netmask, but no check is performed.

# Kinds of netmasks.
NM_DOT = 1
NM_DEC = 2
NM_BIN = 3
NM_UNKNOWN = 0


def is_dot_nm(nm):
    """Return true if the netmask is in dotted decimal notatation."""
    if not is_dot(nm):
        return 0
    # XXX: quite ugly!  Equivalent to:
    # if nm not in map(convert_nm, range(0, 33), (NM_DOT,) * 33, (NM_DEC,)*33):
    # or:
    # if nm not in map(_dec_to_dot_nm, range(0, 33))
    if nm not in ('0.0.0.0', '128.0.0.0', '192.0.0.0', '224.0.0.0',
                    '240.0.0.0', '248.0.0.0', '252.0.0.0', '254.0.0.0',
                    '255.0.0.0', '255.128.0.0', '255.192.0.0', '255.224.0.0',
                    '255.240.0.0', '255.248.0.0', '255.252.0.0', '255.254.0.0',
                    '255.255.0.0', '255.255.128.0', '255.255.192.0',
                    '255.255.224.0', '255.255.240.0', '255.255.248.0',
                    '255.255.252.0', '255.255.254.0', '255.255.255.0',
                    '255.255.255.128', '255.255.255.192', '255.255.255.224',
                    '255.255.255.240', '255.255.255.248', '255.255.255.252',
                    '255.255.255.254', '255.255.255.255'):
        raise ValueError, 'is_dot_nm: not a valid netmask'
    return 1

def is_bin_nm(nm):
    """Return true if the netmask is in binary notatation."""
    if not is_bin(nm):
        return 0
    index = string.find(nm, '0')
    if index != -1:
        if string.find(nm[index:], '1') != -1:
            raise ValueError, 'is_bin_nm: not a netmask'
    return 1

def is_dec_nm(nm):
    """Return true if the netmask is in decimal notatation."""
    if type(nm) not in _StringTypes:
        raise ValueError, 'is_dec_nm: not a netmask'
    try:
        val = long(nm)
    except ValueError:
        return 0
    if val < 0 or val > 32:
        return 0
    return 1


def detect_nm(nm):
    """Return the kind of the netmask.

    Given a netmask, returns the kind of notation (one of
    NM_DOT, NM_DEC, NM_BIN, NM_UNKNOWN).
    """
    if type(nm) is types.IntType:
        nm = repr(nm)
    elif type(nm) is types.LongType:
        nm = repr(nm)[:-1]
    if is_dec_nm(nm):
        return NM_DEC
    elif is_dot_nm(nm):
        return NM_DOT
    elif is_bin_nm(nm):
        return NM_BIN
    return NM_UNKNOWN

def p_detect_nm(nm):
    """Return the kind of the netmask (string).

    Given a netmask, returns a string that represents the
    notation.
    """
    kind = detect_nm(nm)
    if kind == NM_DEC:
        return 'decimal'
    elif kind == NM_DOT:
        return 'dotted decimal'
    elif kind == NM_BIN:
        return 'binary'
    else:
        return 'unknown'

def _dec_to_dot_nm(nm):
    """Decimal to dotted decimal notation conversion."""
    nip = int(nm)
    bin = '1' * nip + '0' * (32 - nip)
    return _dec_to_dot(_bin_to_dec(bin))


def _dot_to_dec_nm(nm):
    """Dotted decimal to decimal notation conversion."""
    bin = _dec_to_bin(_dot_to_dec(nm))
    index = string.find(bin, '0')
    if index == -1:
        return '32'
    return repr(string.find(bin, '0'))



def convert_nm(nm, notation=NM_DOT, iformat=NM_UNKNOWN):
    """Convert among netmask notations.

    Given a netmask, this functions returns the netmask in another
    notation.

    *nm* -- The netmask.

    *notation* -- The format of the output.
    """
    if type(nm) is types.IntType:
        nm = repr(nm)
    elif type(nm) is types.LongType:
        nm = repr(nm)[:-1]
    kind = iformat
    if kind == NM_UNKNOWN:
        kind = detect_nm(nm)
    long = ''
    if kind == NM_DOT:
        long = nm
    elif kind == NM_DEC:
        long = _dec_to_dot_nm(nm)
    elif kind == NM_BIN:
        long = _dec_to_dot(_bin_to_dec(nm))
    if notation == NM_DOT:
        return long
    elif notation == NM_DEC:
        return _dot_to_dec_nm(long)
    elif notation == NM_BIN:
        return _dec_to_bin(_dot_to_dec(long))
    else:
        raise ValueError, 'convert_nm: unknown notation: "' + \
                str(notation) + '"'


class IpNetMask:
    """A CIDR netmask.

    This class represents a CIDR netmask.
    """
    def __init__(self, nm=''):
        if nm:
            self.set(nm)
        else:
            self.nm = ''

    def set(self, nm):
        """Set the netmask."""
        self.nm = convert_nm(nm, notation=detect_nm(nm))

    def get(self):
        """Return the netmask."""
        return self.nm

    def get_dec(self):
        """Return the decimal notation of the netmask."""
        return convert_nm(self.nm, notation=NM_DEC)

    def get_dot(self):
        """Return the dotted decimal notation of the netmask."""
        return convert_nm(self.nm, notation=NM_DOT)

    def get_bin(self):
        """Return the binary notation of the netmask."""
        return convert_nm(self.nm, notation=NM_BIN)


